use crate::{twitter::get_tweet_media, ScootalooError};

use std::{borrow::Cow, error::Error};

use egg_mode::tweet::Tweet;

use elefren::prelude::*;

use log::{error, info, warn};

use reqwest::Url;

use tokio::{
    fs::{create_dir_all, remove_file, File},
    io::copy,
};

/// Generate associative table between media ids and tweet extended entities
pub async fn generate_media_ids(
    tweet: &Tweet,
    cache_path: &str,
    mastodon: &Mastodon,
) -> (String, Vec<String>) {
    let mut media_url = "".to_string();
    let mut media_ids: Vec<String> = vec![];

    if let Some(m) = &tweet.extended_entities {
        // create tasks list
        let mut tasks = vec![];

        // size of media_ids vector, should be equal to the media vector
        media_ids.resize(m.media.len(), String::new());

        info!("{} medias in tweet", m.media.len());

        for (i, media) in m.media.iter().enumerate() {
            // attribute media url
            media_url = media.url.clone();

            // clone everything we need
            let cache_path = String::from(cache_path);
            let media = media.clone();
            let mastodon = mastodon.clone();

            let task = tokio::task::spawn(async move {
                info!("Start treating {}", media.media_url_https);
                // get the tweet embedded media
                let local_tweet_media_path = match get_tweet_media(&media, &cache_path).await {
                    Ok(l) => l,
                    Err(e) => {
                        return Err(ScootalooError::new(&format!(
                            "Cannot get tweet media for {}: {}",
                            &media.url, e
                        )))
                    }
                };

                // upload media to Mastodon
                let mastodon_media = mastodon.media(Cow::from(local_tweet_media_path.to_owned()));
                // at this point, we can safely erase the original file
                // it doesn’t matter if we can’t remove, cache_media fn is idempotent
                remove_file(&local_tweet_media_path).await.ok();

                let mastodon_media = match mastodon_media {
                    Ok(m) => m,
                    Err(e) => {
                        return Err(ScootalooError::new(&format!(
                            "Attachment {} cannot be uploaded to Mastodon Instance: {}",
                            &local_tweet_media_path, e
                        )))
                    }
                };

                Ok((i, mastodon_media.id))
            });

            tasks.push(task);
        }

        for task in tasks {
            match task.await {
                // insert the media at the right place
                Ok(Ok((i, v))) => media_ids[i] = v,
                Ok(Err(e)) => warn!("{}", e),
                Err(e) => error!("Something went wrong when joining the main thread: {}", e),
            }
        }
    } else {
        info!("No media in tweet");
    }

    // in case some media_ids slot remained empty due to errors, remove them
    media_ids.retain(|x| !x.is_empty());

    (media_url, media_ids)
}

/// Gets and caches Twitter Media inside the determined temp dir
pub async fn cache_media(u: &str, t: &str) -> Result<String, Box<dyn Error>> {
    // create dir
    create_dir_all(t).await?;

    // get file
    let mut response = reqwest::get(u).await?;

    // create local file
    let url = Url::parse(u)?;
    let dest_filename = url
        .path_segments()
        .ok_or_else(|| {
            ScootalooError::new(&format!(
                "Cannot determine the destination filename for {}",
                u
            ))
        })?
        .last()
        .ok_or_else(|| {
            ScootalooError::new(&format!(
                "Cannot determine the destination filename for {}",
                u
            ))
        })?;

    let dest_filepath = format!("{}/{}", t, dest_filename);

    let mut dest_file = File::create(&dest_filepath).await?;

    while let Some(chunk) = response.chunk().await? {
        copy(&mut &*chunk, &mut dest_file).await?;
    }

    Ok(dest_filepath)
}

#[cfg(test)]
mod tests {
    use super::*;

    use std::{fs::remove_dir_all, path::Path};

    const TMP_DIR: &'static str = "/tmp/scootaloo_test";

    #[tokio::test]
    async fn test_cache_media() {
        let dest = cache_media(
            "https://forum.nintendojo.fr/styles/prosilver/theme/images/ndfr_casual.png",
            TMP_DIR,
        )
        .await
        .unwrap();

        assert!(Path::new(&dest).exists());

        remove_dir_all(TMP_DIR).unwrap();
    }
}
